" vis.vim:
" Function: Perform an Ex command on a visual highlighted block (CTRL-V).
"
" Requires: Requires 6.0 or later  (this script is a plugin)
"
" Usage:    Mark visual block (CTRL-V) or visual character (v),
"           press ':B ' and enter an Ex command [cmd].
"
"           ex. Use ctrl-v to visually mark the block then use
"                 :B cmd     (will appear as   :'<,'>B cmd )
"
"           ex. Use v to visually mark the block then use
"                 :B cmd     (will appear as   :'<,'>B cmd )
"
"           Command-line completion is supported for Ex commands.
"
" Note:     There must be a space before the '!' when invoking external shell
"           commands, eg. ':B !sort'. Otherwise an error is reported.
"
" Author:   Charles E. Campbell <cec@NdjOinnSi.gPsfAc.nMasa.gov>, remove NOSPAM first
"           Based on idea of Stefan Roemer <roemer@informatik.tu-muenchen.de>
"
" Version:  5
"
" History:
"    5 : Jun 21, 2002 - now supports character-visual mode (v) and
"                       linewise-visual mode (V)
"    4 : Jun 20, 2002 - saves sol, sets nosol, restores sol
"                     - bugfix: 0 inserted: 0\<c-v>G$\"ad
"                     - Fixed double-loading (was commented
"                       out for debugging purposes)
"    3 : Jun 20, 2002 - saves ve, unsets ve, restores ve
"    2 : Jun 19, 2002 - Charles Campbell's <vis.vim>
"    1 : epoch        - Stefan Roemer's    <vis.vim>
" ------------------------------------------------------------------------------

" Exit quickly when <Vis.vim> has already been loaded or when 'compatible' is set
if exists("loaded_visvim") || &cp
  finish
endif
let loaded_visvim= 1

" ------------------------------------------------------------------------------

"  -range       : VisBlockCmd operates on the range itself
"  -com=command : Ex command and arguments
"  -nargs=+     : arguments may be supplied, up to any quantity
com! -range -nargs=+ -com=command B <line1>,<line2>call s:VisBlockCmd('<args>')

" ------------------------------------------------------------------------------

" VisBlockCmd:
fu! <SID>VisBlockCmd(cmd) range

  " retain and re-use same visual mode
  let vmode= visualmode()

  " save options which otherwise may interfere
  let keep_ve= &ve
  set ve=
  let keep_sol= &sol
  set nosol
  let keep_lz= &lz
  set lz

  " Save any contents in register a
  let rega= @a

  if vmode == 'V'
   exe "'<,'>".a:cmd
  else

   " Initialize so that begline<endline and begcol<endcol.
   let begcol   = virtcol("'<")-1
   let endcol   = virtcol("'>")-1
   if begcol > endcol
    let begcol  = virtcol("'>")-1
    let endcol  = virtcol("'<")-1
   endif
   let begline  = a:firstline
   let endline  = a:lastline
   if begline > endline
    let begline = a:lastline
    let endline = a:firstline
   endif
"  call Decho('beg['.begline.','.begcol.'] end['.endline.','.endcol.'] ragged='.ragged)

   " Modify Selected Region:
   " =======================
   " delete selected region into register "a
   " put cut-out text at end-of-file
   " apply command to those lines
   " visual-block select the modified text in those lines
   " delete excess lines
   " put modified text back into file
   norm! gv"ad
   $
   pu_
   let lastline= line("$")
   silent norm! "ap
   exe '.,$'.a:cmd
   exe lastline
   exe "norm! 0".vmode."G$\"ad"
   silent exe lastline.',$d'
   exe begline
   if begcol > 1
    exe 'norm! '.begcol."\<bar>\"ap"
   else
    norm! 0"aP
   endif

   " attempt to restore gv -- this is limited, it will
   " select the same size region in the same place as before,
   " not necessarily the changed region
   let begcol= begcol+1
   let endcol= endcol+1
   silent exe begline
   silent exe 'norm! '.begcol."\<bar>".vmode
   silent exe endline
   silent exe 'norm! '.endcol."\<bar>\<esc>"
   silent exe begline
   silent exe 'norm! '.begcol."\<bar>"
  endif

  " restore register a and options
  let @a  = rega
  let &ve = keep_ve
  let &sol= keep_sol
  let &lz = keep_lz
  
endfunction

" ------------------------------------------------------------------------------
