" -*- vim -*-
" $Id: bob-cvs.vim,v 1.6 2001/09/21 19:17:09 rhiestan Exp $
"
" Vim global plugin to assist in working with CVS-controlled files.
" Last Change: $Date: 2001/09/21 19:17:09 $
" Maintainer: Robert Hiestand <robert.hiestand@mindspring.com>
"
" Provides functions to invoke various CVS commands on the current file.
" The output of the commands is captured in a new scratch window.  For
" convenience, if the functions are invoked on a CVS output window, the
" original file is used for the cvs operation instead after the window is
" split.  This is primarily useful when running CVSCommit and you need to see
" the changes made, so that CVSDiff is usable and shows up in another window.
"
" These functions are exported into the global environment, meaning they are
" directly accessed without prepending '<PLUG>'.  This is because the author
" directly accesses several of them without using the mappings in order to
" pass parameters.
"
" Several of these act immediately, such as
"
" CVSAdd        Performs "cvs add" on the current file.
"
" CVSAnnotate   Performs "cvs annotate" on the current file.  If not given an
"               argument, uses the most recent version of the file on the current
"               branch.  Otherwise, the argument is used as a revision number.
"
" CVSCommit     This is a two-stage command.  The first step opens a buffer to
"               accept a log message.  When that buffer is written, it is
"               automatically closed and the file is committed using the
"               information from that log message.  If the file should not be
"               committed, just destroy the log message buffer without writing
"               it.
"
" CVSDiff       With no arguments, this performs "cvs diff" on the current
"               file.  With one argument, "cvs diff" is performed on the
"               current file against the specified revision.  With two
"               arguments, cvs diff is performed between the specified
"               revisions of the current file.  This command uses the
"               'cvsdiffopt' variable to specify diff options.  If that
"               variable does not exist, then 'wbBc' is assumed.  If you wish
"               to have no options, then set it to the empty string.
"
" CVSLog        Performs "cvs log" on the current file.
"
" CVSStatus     Performs "cvs status" on the current file.
"
" CVSUpdate     Performs "cvs update" on the current file.
"
" CVSReview     Retrieves a particular version of the current file.  If no
"               argument is given, the most recent version of the file on
"               the current branch is retrieved.  The specified revision is
"               retrieved into a new buffer.
"
" CVSVimDiff    With no arguments, this prompts the user for a revision and
"               then uses vimdiff to display the differences between the current
"               file and the specified revision.  If no revision is specified,
"               the most recent version of the file on the current branch is used.
"               With one argument, that argument is used as the revision as
"               above.  With two arguments, the differences between the two
"               revisions is displayed using vimdiff.
"
" By default, a mapping is defined for each command.  User-provided mappings
" can be used instead by mapping to <Plug>CommandName, for instance:
"
" nnoremap ,ca <Plug>CVSAdd
"
" The default mappings are as follow:
"
"   <Leader>ca CVSAdd
"   <Leader>cn CVSAnnotate
"   <Leader>cc CVSCommit
"   <Leader>cd CVSDiff
"   <Leader>cl CVSLog
"   <Leader>cr CVSReview
"   <Leader>cs CVSStatus
"   <Leader>cu CVSUpdate
"   <Leader>cv CVSVimDiff

if exists("loaded_bobcvs")
   finish
endif
let loaded_bobcvs = 1

""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"                               Utility functions                            "
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

" Go to the directory in which the current CVS-controlled file is located.
" If this is a CVS command buffer, first switch to the original file.

function! <SID>CVSChangeToCurrentFileDir()
  let oldCwd=getcwd()
  let newCwd=expand("%:h")
  if strlen(newCwd) > 0
    execute 'cd' escape(newCwd, ' ')
  endif
  return oldCwd
endfunction

" Creates a new scratch buffer and captures the output from execution of the
" given command.  The name of the scratch buffer is returned.

function! <SID>CVSCreateCommandBuffer(cmd, cmdname, filename)
  let origBuffNR=bufnr("%")
  let bufName='*' . a:cmdname . '*' . ' ' . a:filename
  let counter=0
  let currentBufName = bufName
  while buflisted(currentBufName)
    let counter=counter + 1
    let currentBufName=bufName . ' (' . counter . ')'
  endwhile
  let currentBufName = escape(currentBufName, ' *\')
  execute 'edit' currentBufName
  set buftype=nofile
  set noswapfile
  set filetype=
  let b:cvsOrigBuffNR=origBuffNR
  silent execute a:cmd
  $d
  1
  return currentBufName
endfunction

function! <SID>CVSBufferCheck()
  if exists("b:cvsOrigBuffNR")
    if bufexists(b:cvsOrigBuffNR)
      " Why is this failing?
      execute "sbuffer" . b:cvsOrigBuffNR
      return 1
      "return b:cvsOrigBuffNR
    else
      " Original buffer no longer exists.
      return -1 
    endif
  else
    " No original buffer
    return 0
  endif
endfunction

""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
"                                 CVS functions                              "
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

function! <SID>CVSCommit()
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck ==  -1
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  let originalBuffer=bufnr("%")
  let messageFileName = escape(tempname().'*log message*', ' ?*\')
  let newCwd=expand("%:h")
  let fileName=expand("%:t")

  execute 'au BufWritePost' messageFileName 'call <SID>CVSFinishCommit("' . messageFileName . '", "' . newCwd . '", "' . fileName . '") | au! * ' messageFileName
  execute 'au BufDelete' messageFileName 'au! * ' messageFileName

  execute 'edit' messageFileName
  let b:cvsOrigBuffNR=originalBuffer
endfunction

function! <SID>CVSFinishCommit(messageFile, targetDir, targetFile)
  if filereadable(a:messageFile)
    let oldCwd=getcwd()
    if strlen(a:targetDir) > 0
      execute 'cd' escape(a:targetDir, ' ')
    endif
    let resultBufferName=<SID>CVSCreateCommandBuffer('0r!cvs commit -F "' . a:messageFile . '" "'. a:targetFile . '"', 'cvscommit', expand("%"))
    execute 'cd' escape(oldCwd, ' ')
    execute 'bw' escape(a:messageFile, ' *?\')
    silent execute '!rm' a:messageFile
    return resultBufferName
  else
    echo "Can't read message file; no commit is possible."
    return ""
  endif
endfunction

function! <SID>CVSUpdate()
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck == -1 
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  let oldCwd=<SID>CVSChangeToCurrentFileDir()
  let resultBufferName=<SID>CVSCreateCommandBuffer('0r!cvs update "' . expand("%") . '"', 'cvsupdate', expand("%"))
  execute 'cd' escape(oldCwd, ' ')
  return resultBufferName
endfunction

function! <SID>CVSDiff(...)
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck == -1 
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  if a:0 == 1
    let revOptions = '-r' . a:1
    let caption = 'cvsdiff ' . a:1 . ' -> current'
  elseif a:0 == 2
    let revOptions = '-r' . a:1 . ' -r' . a:2
    let caption = 'cvsdiff ' . a:1 . ' -> ' . a:2
  else
    let revOptions = ''
    let caption = 'cvsdiff'
  endif

  let diffoptionstring=' -wbBc '

  let cvsdiffopt="-1"
  if exists("g:cvsdiffopt")
    let cvsdiffopt=g:cvsdiffopt
  elseif exists("b:cvsdiffopn")
    let cvsdiffopt=b:cvsdiffopt
  elseif exists("w:cvsdiffopt")
    let cvsdiffopt=w:cvsdiffopt
  endif

  if cvsdiffopt != "-1"
    if cvsdiffopt == ""
      let diffoptionstring=""
    else
      let diffoptionstring=" -" . cvsdiffopt . " "
    endif
  endif

  let oldCwd=<SID>CVSChangeToCurrentFileDir()
  let resultBufferName=<SID>CVSCreateCommandBuffer('0r!cvs diff ' . diffoptionstring . revOptions . ' "' . expand("%") . '"', caption, expand("%"))
  "call SetSyn("diff")
  set syntax=diff
  execute 'cd' escape(oldCwd, ' ')
  return resultBufferName
endfunction

function! <SID>CVSLog()
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck == -1 
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  let oldCwd=<SID>CVSChangeToCurrentFileDir()
  let resultBufferName=<SID>CVSCreateCommandBuffer('0r!cvs log "' . expand("%") . '"', 'cvslog', expand("%"))
  execute 'cd' escape(oldCwd, ' ')
  return resultBufferName
endfunction

function! <SID>CVSAdd()
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck == -1 
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  let oldCwd=<SID>CVSChangeToCurrentFileDir()
  let resultBufferName=<SID>CVSCreateCommandBuffer('0r!cvs add "' . expand("%") . '"', 'cvsadd', expand("%"))
  execute 'cd' escape(oldCwd, ' ')
  return resultBufferName
endfunction

function! <SID>CVSAnnotate(...)
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck == -1 
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  let currentLine=line(".")
  let oldCwd=<SID>CVSChangeToCurrentFileDir()

  if a:0 == 0
    let revision=system("cvs status " . escape(expand("%"), " *?\\"))
    if(v:shell_error)
      echo "Unable to obtain status for " . expand("%")
      return
    endif
    let revision=substitute(revision, '^\_.*Working revision:\s*\(\d\+\%(\.\d\+\)\+\)\_.*$', '\1', "")
  else
    let revision=a:1
  endif

  let resultBufferName=<SID>CVSCreateCommandBuffer('0r!cvs -q annotate -r ' . revision . ' "' . expand("%") . '" 2>/dev/null', 'cvsannotate', expand("%"))
  exec currentLine
  execute 'cd' escape(oldCwd, ' ')
  return resultBufferName
endfunction

function! <SID>CVSReview(...)
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck == -1
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  if a:0 == 0
    let versiontag="current"
    let versionOption=""
  else
    let versiontag=a:1
    let versionOption=" -r " . versiontag . " "
  endif

  let origFileType=&filetype
  let oldCwd=<SID>CVSChangeToCurrentFileDir()
  let resultBufferName=<SID>CVSCreateCommandBuffer('0r!cvs update -p' . versionOption . ' "' . expand("%") . '" 2>/dev/null', 'cvsreview -- ' . versiontag, expand("%"))
  execute 'cd' escape(oldCwd, ' ')
  let &filetype=origFileType
  return resultBufferName
endfunction

function! <SID>CVSStatus()
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck == -1 
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  let oldCwd=<SID>CVSChangeToCurrentFileDir()
  let resultBufferName=<SID>CVSCreateCommandBuffer('0r!cvs status "' . expand("%") . '" 2>/dev/null', 'cvsstatus', expand("%"))
  execute 'cd' escape(oldCwd, ' ')
  return resultBufferName
endfunction

function! <SID>CVSVimDiff(...)
  let cvsBufferCheck=<SID>CVSBufferCheck()
  if cvsBufferCheck == -1 
    echo "Original buffer no longer exists, aborting."
    return ""
  endif

  let originalBuffer=bufnr("%")

  let isUsingSource=1

  if(a:0 == 0)
    let resultBufferName=<SID>CVSReview()
  elseif(a:0 == 1)
    let resultBufferName=<SID>CVSReview(a:1)
  else
    let isUsingSource=0
    let resultBufferName=<SID>CVSReview(a:1)
    set buftype=
    w!
    execute "au BufDelete" resultBufferName "call delete(\"".resultBufferName ."\") | au! BufDelete" resultBufferName
    execute "buffer" originalBuffer
    let originalBuffer=bufnr(resultBufferName)
    let resultBufferName=<SID>CVSReview(a:2)
  endif

  if isUsingSource
    " Provide for restoring state of original buffer
    execute "au BufDelete" resultBufferName "call setbufvar(" b:cvsOrigBuffNR ", \"&diff\", 0)"
    execute "au BufDelete" resultBufferName "call setbufvar(" b:cvsOrigBuffNR ", \"&foldcolumn\", 0)"
    execute "au BufDelete" resultBufferName "au! BufDelete " escape(resultBufferName, ' *?\')
  endif

  execute "silent leftabove vertical diffsplit" escape(bufname(originalBuffer), ' *?\')
  execute bufwinnr(originalBuffer) "wincmd w"
  return resultBufferName
endfunction

com! CVSCommit call <SID>CVSCommit()
com! CVSUpdate call <SID>CVSUpdate()
com! -nargs=* CVSDiff call <SID>CVSDiff(<f-args>)
com! CVSLog call <SID>CVSLog()
com! CVSAdd call <SID>CVSAdd()
com! -nargs=? CVSAnnotate call <SID>CVSAnnotate(<f-args>)
com! -nargs=? CVSReview call <SID>CVSReview(<f-args>)
com! CVSStatus call <SID>CVSStatus()
com! -nargs=* CVSVimDiff call <SID>CVSVimDiff(<f-args>)

if !hasmapto('<Plug>CVSAdd')
  nmap <unique> <Leader>ca <Plug>CVSAdd
endif
nnoremap <unique> <Plug>CVSAdd :CVSAdd<CR>
if !hasmapto('<Plug>CVSCommit')
  nmap <unique> <Leader>cc <Plug>CVSCommit
endif
nnoremap <unique> <Plug>CVSCommit :CVSCommit<CR>
if !hasmapto('<Plug>CVSDiff')
  nmap <unique> <Leader>cd <Plug>CVSDiff
endif
nnoremap <unique> <Plug>CVSDiff :CVSDiff<CR>
if !hasmapto('<Plug>CVSLog')
  nmap <unique> <Leader>cl <Plug>CVSLog
endif
nnoremap <unique> <Plug>CVSLog :CVSLog<CR>
if !hasmapto('<Plug>CVSAnnotate')
  nmap <unique> <Leader>cn <Plug>CVSAnnotate
endif
nnoremap <unique> <Plug>CVSAnnotate :CVSAnnotate<CR>
if !hasmapto('<Plug>CVSReview')
  nmap <unique> <Leader>cr <Plug>CVSReview
endif
nnoremap <unique> <Plug>CVSReview :CVSReview<CR>
if !hasmapto('<Plug>CVSStatus')
  nmap <unique> <Leader>cs <Plug>CVSStatus
endif
nnoremap <unique> <Plug>CVSStatus :CVSStatus<CR>
if !hasmapto('<Plug>CVSUpdate')
  nmap <unique> <Leader>cu <Plug>CVSUpdate
endif
nnoremap <unique> <Plug>CVSUpdate :CVSUpdate<CR>
if !hasmapto('<Plug>CVSVimDiff')
  nmap <unique> <Leader>cv <Plug>CVSVimDiff
endif
nnoremap <unique> <Plug>CVSVimDiff :CVSVimDiff<CR>
